// insert header here
// mbm. gpl.

#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/if_arp.h>
#include <asm/uaccess.h>
#include <linux/wireless.h>

#include <net/iw_handler.h>
#include <wlioctl.h>

#define DEBUG

static struct net_device *dev;

/* The frequency of each channel in MHz */
const long channel_frequency[] = {
	2412, 2417, 2422, 2427, 2432, 2437, 2442,
	2447, 2452, 2457, 2462, 2467, 2472, 2484
};
#define NUM_CHANNELS ( sizeof(channel_frequency) / sizeof(channel_frequency[0]) )

static int wl_ioctl(struct net_device *dev, int cmd, void *buf, int len)
{
	mm_segment_t old_fs = get_fs();
	struct ifreq ifr;
	int ret;
	wl_ioctl_t ioc;
	ioc.cmd = cmd;
	ioc.buf = buf;
	ioc.len = len;
	strncpy(ifr.ifr_name, dev->name, IFNAMSIZ);
	ifr.ifr_data = (caddr_t) &ioc;
	set_fs(KERNEL_DS);
	ret = dev->do_ioctl(dev,&ifr,SIOCDEVPRIVATE);
	set_fs (old_fs);
	return ret;
}

static int wlcompat_ioctl_getiwrange(struct net_device *dev,
				    char *extra)
{
	int i, k;
	struct iw_range *range;

	range = (struct iw_range *) extra;

	range->we_version_compiled = WIRELESS_EXT;
	range->we_version_source = WIRELESS_EXT;
	
	range->min_nwid = range->max_nwid = 0;
	
	range->num_channels = NUM_CHANNELS;
	k = 0;
	for (i = 0; i < NUM_CHANNELS; i++) {
		range->freq[k].i = i + 1;
		range->freq[k].m = channel_frequency[i] * 100000;
		range->freq[k].e = 1;
		k++;
		if (k >= IW_MAX_FREQUENCIES)
			break;
	}
	range->num_frequency = k;
	range->sensitivity = 3;

	/* nbd: don't know what this means, but other drivers set it this way */
	range->pmp_flags = IW_POWER_PERIOD;
	range->pmt_flags = IW_POWER_TIMEOUT;
	range->pm_capa = IW_POWER_PERIOD | IW_POWER_TIMEOUT | IW_POWER_UNICAST_R;

	range->min_rts = 0;
	range->max_rts = 2347;
	range->min_frag = 256;
	range->max_frag = 2346;

	range->min_pmp = 0;
	range->max_pmp = 65535000;
	range->min_pmt = 0;
	range->max_pmt = 65535 * 1000;

	range->txpower_capa = IW_TXPOW_MWATT;

	return 0;
}

static int wlcompat_ioctl(struct net_device *dev,
			 struct iw_request_info *info,
			 union iwreq_data *wrqu,
			 char *extra)
{
	int err = 0;
	
	switch (info->cmd) {
		case SIOCGIWNAME:
			strcpy(wrqu->name, "IEEE 802.11-DS");
			break;
		case SIOCGIWFREQ:
		{
			channel_info_t ci;
			wl_ioctl(dev,WLC_GET_CHANNEL, &ci, sizeof(ci));
			wrqu->freq.m = ci.target_channel;
			wrqu->freq.e = 0;
			break;
		}
		case SIOCSIWFREQ:
		{
			if (wrqu->freq.e == 1) {
				int channel = 0;
				int f = wrqu->freq.m / 100000;
				while ((channel < NUM_CHANNELS + 1) && (f != channel_frequency[channel]))
					channel++;
				
				if (channel == NUM_CHANNELS) { // channel not found
					err = -EINVAL;
				} else {
					wrqu->freq.e = 0;
					wrqu->freq.m = channel + 1;
				}
			}
			if ((wrqu->freq.e == 0) && (wrqu->freq.m < 1000)) {
				wl_ioctl(dev, WLC_SET_CHANNEL, &wrqu->freq.m, sizeof(int));
			} else {
				err = -EINVAL;
			}
		}
		case SIOCGIWAP:
		{
			wrqu->ap_addr.sa_family = ARPHRD_ETHER;
			wl_ioctl(dev,WLC_GET_BSSID,wrqu->ap_addr.sa_data,6);
			break;
		}
		case SIOCGIWESSID:
		{
			wlc_ssid_t ssid;
			wl_ioctl(dev,WLC_GET_SSID, &ssid, sizeof(wlc_ssid_t));
			wrqu->essid.flags = wrqu->data.flags = 1;
			wrqu->essid.length = wrqu->data.length = ssid.SSID_len + 1;
			memcpy(extra,ssid.SSID,ssid.SSID_len + 1);
			break;
		}
		case SIOCSIWESSID:
		{
			wlc_ssid_t ssid;
			memset(&ssid, 0, sizeof(ssid));
			ssid.SSID_len = strlen(extra);
			if (ssid.SSID_len > WLC_ESSID_MAX_SIZE)
				ssid.SSID_len = WLC_ESSID_MAX_SIZE;
			memcpy(ssid.SSID, extra, ssid.SSID_len);
			wl_ioctl(dev, WLC_SET_SSID, &ssid, sizeof(ssid));
			break;
		}
		case SIOCGIWRTS:
		{
			wl_ioctl(dev,WLC_GET_RTS,&(wrqu->rts.value),sizeof(int));
			break;
		}
		case SIOCGIWFRAG:
		{
			wl_ioctl(dev,WLC_GET_FRAG,&(wrqu->frag.value),sizeof(int));
			break;
		}
		case SIOCGIWTXPOW:
		{
			wrqu->txpower.value = 0;
			wl_ioctl(dev,WLC_GET_TXPWR, &(wrqu->txpower.value), sizeof(int));
			wrqu->txpower.fixed = 0;
			wrqu->txpower.disabled = 0;
			wrqu->txpower.flags = IW_TXPOW_MWATT;
			break;
		}
		case SIOCGIWENCODE:
		{
			wrqu->data.flags = IW_ENCODE_DISABLED;
			break;
		}
		case SIOCGIWRANGE:
		{
			err = wlcompat_ioctl_getiwrange(dev, extra);
			break;
		}
		default:
		{
			err = -EINVAL;
			break;
		}
	}
	
	return err;
}

static const iw_handler	 wlcompat_handler[] = {
	NULL,			/* SIOCSIWCOMMIT */
	wlcompat_ioctl,		/* SIOCGIWNAME */
	NULL,			/* SIOCSIWNWID */
	NULL,			/* SIOCGIWNWID */
	wlcompat_ioctl,		/* SIOCSIWFREQ */
	wlcompat_ioctl,		/* SIOCGIWFREQ */
	NULL,			/* SIOCSIWMODE */
	NULL,			/* SIOCGIWMODE */
	NULL,			/* SIOCSIWSENS */
	NULL,			/* SIOCGIWSENS */
	NULL,			/* SIOCSIWRANGE */
	wlcompat_ioctl,		/* SIOCGIWRANGE */
	NULL,			/* SIOCSIWPRIV */
	NULL,			/* SIOCGIWPRIV */
	NULL,			/* SIOCSIWSTATS */
	NULL,			/* SIOCGIWSTATS */
	iw_handler_set_spy,	/* SIOCSIWSPY */
	iw_handler_get_spy,	/* SIOCGIWSPY */
	iw_handler_set_thrspy,  /* SIOCSIWTHRSPY */
	iw_handler_get_thrspy,  /* SIOCGIWTHRSPY */
	NULL,			/* SIOCSIWAP */
	wlcompat_ioctl,		/* SIOCGIWAP */
	NULL,			/* -- hole -- */
	NULL,			/* SIOCGIWAPLIST */
	NULL,			/* -- hole -- */
	NULL,			/* -- hole -- */
	wlcompat_ioctl,		/* SIOCSIWESSID */
	wlcompat_ioctl,		/* SIOCGIWESSID */
	NULL,			/* SIOCSIWNICKN */
	NULL,			/* SIOCGIWNICKN */
	NULL,			/* -- hole -- */
	NULL,			/* -- hole -- */
	NULL,			/* SIOCSIWRATE */
	NULL,			/* SIOCGIWRATE */
	NULL,			/* SIOCSIWRTS */
	wlcompat_ioctl,		/* SIOCGIWRTS */
	NULL,			/* SIOCSIWFRAG */
	wlcompat_ioctl,		/* SIOCGIWFRAG */
	NULL,			/* SIOCSIWTXPOW */
	wlcompat_ioctl,		/* SIOCGIWTXPOW */
	NULL,			/* SIOCSIWRETRY */
	NULL,			/* SIOCGIWRETRY */
	NULL,			/* SIOCSIWENCODE */
	wlcompat_ioctl,		/* SIOCGIWENCODE */
};

static const struct iw_handler_def      wlcompat_handler_def =
{
	.standard	= (iw_handler *) wlcompat_handler,
	.num_standard	= sizeof(wlcompat_handler)/sizeof(iw_handler),
	.private	= NULL,
	.num_private	= 0,
	.private_args	= NULL, 
	.num_private_args = 0,
};

#ifdef DEBUG
static int (*old_ioctl)(struct net_device *dev, struct ifreq *ifr, int cmd);
static int new_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd) {
	int ret = old_ioctl(dev,ifr,cmd);
	printk("dev: %s ioctl: 0x%04x\n",dev->name,cmd);
	if (cmd==SIOCDEVPRIVATE) {
		int x;
		wl_ioctl_t *ioc = (wl_ioctl_t *)ifr->ifr_data;
		unsigned char *buf = ioc->buf;
		printk("   cmd: %d buf: 0x%08x len: %d\n",ioc->cmd,&(ioc->buf),ioc->len);
		printk("   ->");
		for (x=0;x<ioc->len && x<128 ;x++) {
			printk("%02X",buf[x]);
		}
		printk("\n");
	}
	return ret;
}
#endif

static int __init wlcompat_init()
{
	dev = dev_get_by_name("eth1");
#ifdef DEBUG
	old_ioctl = dev->do_ioctl;
	dev->do_ioctl = new_ioctl;
#endif
	dev->wireless_handlers = (struct iw_handler_def *)&wlcompat_handler_def;
	return 0;
}

static void __exit wlcompat_exit()
{
	dev->wireless_handlers = NULL;
#ifdef DEBUG
	dev->do_ioctl = old_ioctl;
#endif
	return;
}

EXPORT_NO_SYMBOLS;
MODULE_AUTHOR("openwrt.org");
MODULE_LICENSE("GPL");

module_init(wlcompat_init);
module_exit(wlcompat_exit);
